@@ -3,38 +3,40 @@
 import os
 import pathlib
 from ctypes import *
+from typing import Tuple
+
 from . import sharedLibraryExtension, platform_tuple
 from .fmi1 import _FMU, printLogMessage
 
 
-fmi3Instance             = c_void_p
-fmi3InstanceEnvironment  = c_void_p
-fmi3FMUState             = c_void_p
-fmi3ValueReference       = c_uint
-fmi3Float32              = c_float
-fmi3Float64              = c_double
-fmi3Int8                 = c_int8
-fmi3UInt8                = c_uint8
-fmi3Int16                = c_int16
-fmi3UInt16               = c_uint16
-fmi3Int32                = c_int32
-fmi3UInt32               = c_uint32
-fmi3Int64                = c_int64
-fmi3UInt64               = c_uint64
-fmi3Boolean              = c_bool
-fmi3Char                 = c_char
-fmi3String               = c_char_p
-fmi3Byte                 = c_char
-fmi3Binary               = c_char_p
-fmi3Clock                = c_bool
+fmi3Instance            = c_void_p
+fmi3InstanceEnvironment = c_void_p
+fmi3FMUState            = c_void_p
+fmi3ValueReference      = c_uint
+fmi3Float32             = c_float
+fmi3Float64             = c_double
+fmi3Int8                = c_int8
+fmi3UInt8               = c_uint8
+fmi3Int16               = c_int16
+fmi3UInt16              = c_uint16
+fmi3Int32               = c_int32
+fmi3UInt32              = c_uint32
+fmi3Int64               = c_int64
+fmi3UInt64              = c_uint64
+fmi3Boolean             = c_bool
+fmi3Char                = c_char
+fmi3String              = c_char_p
+fmi3Byte                = c_char
+fmi3Binary              = c_char_p
+fmi3Clock               = c_bool
 
 # values for fmi3Boolean
-fmi3True  = 1
-fmi3False = 0
+fmi3True  = c_bool(True)
+fmi3False = c_bool(False)
 
 # values for fmi3Clock
-fmi3ClockActive   = 1
-fmi3ClockInactive = 0
+fmi3ClockActive   = c_bool(True)
+fmi3ClockInactive = c_bool(False)
 
 # enum fmi3Status
 fmi3Status  = c_int
@@ -66,19 +68,30 @@ fmi3CallbackLockPreemptionTYPE     = CFUNCTYPE(None)
 fmi3CallbackUnlockPreemptionTYPE   = CFUNCTYPE(None)
 
 
-def intermediateUpdate(*args):
+def intermediateUpdate(instanceEnvironment: fmi3InstanceEnvironment,
+                       intermediateUpdateTime: fmi3Float64,
+                       clocksTicked: fmi3Boolean,
+                       intermediateVariableSetRequested: fmi3Boolean,
+                       intermediateVariableGetAllowed: fmi3Boolean,
+                       intermediateStepFinished: fmi3Boolean,
+                       canReturnEarly: fmi3Boolean,
+                       earlyReturnRequested: POINTER(fmi3Boolean),
+                       earlyReturnTime: POINTER(fmi3Float64)) -> fmi3Status:
+
+    earlyReturnRequested.contents = fmi3False
+
     return fmi3OK
 
 
-def printLogMessage(instanceEnvironment, instanceName, status, category, message):
+def printLogMessage(instanceEnvironment: fmi3InstanceEnvironment,
+                    instanceName: fmi3String,
+                    status: fmi3Status,
+                    category: fmi3Status,
+                    message: fmi3Status) -> None:
     """ Print the FMU's log messages to the command line """
 
     label = ['OK', 'WARNING', 'DISCARD', 'ERROR', 'FATAL', 'PENDING'][status]
-    print("[%s] %s" % (label, message))
-
-
-def stepFinished(instanceEnvironment, status):
-    pass
+    print(f"[{label}] {message.decode('utf-8')}")
 
 
 class _FMU3(_FMU):
@@ -105,7 +118,7 @@ class _FMU3(_FMU):
         self._fmi3Function('fmi3InstantiateModelExchange', [
             (fmi3String,                         'instanceName'),
             (fmi3String,                         'instantiationToken'),
-            (fmi3String,                         'resourceLocation'),
+            (fmi3String,                         'resourcePath'),
             (fmi3Boolean,                        'visible'),
             (fmi3Boolean,                        'loggingOn'),
             (fmi3InstanceEnvironment,            'instanceEnvironment'),
@@ -115,7 +128,7 @@ class _FMU3(_FMU):
         self._fmi3Function('fmi3InstantiateCoSimulation', [
             (fmi3String,                         'instanceName'),
             (fmi3String,                         'instantiationToken'),
-            (fmi3String,                         'resourceLocation'),
+            (fmi3String,                         'resourcePath'),
             (fmi3Boolean,                        'visible'),
             (fmi3Boolean,                        'loggingOn'),
             (fmi3Boolean,                        'eventModeUsed'),
@@ -128,18 +141,18 @@ class _FMU3(_FMU):
         ], fmi3Instance)
 
         self._fmi3Function('fmi3InstantiateScheduledExecution', [
-            (fmi3String, 'instanceName'),
-            (fmi3String, 'instantiationToken'),
-            (fmi3String, 'resourceLocation'),
-            (fmi3Boolean, 'visible'),
-            (fmi3Boolean, 'loggingOn'),
-            (POINTER(fmi3ValueReference), 'requiredIntermediateVariables'),
-            (c_size_t, 'nRequiredIntermediateVariables'),
-            (fmi3InstanceEnvironment, 'instanceEnvironment'),
-            (fmi3CallbackLogMessageTYPE, 'logMessage'),
+            (fmi3String,                         'instanceName'),
+            (fmi3String,                         'instantiationToken'),
+            (fmi3String,                         'resourcePath'),
+            (fmi3Boolean,                        'visible'),
+            (fmi3Boolean,                        'loggingOn'),
+            (POINTER(fmi3ValueReference),        'requiredIntermediateVariables'),
+            (c_size_t,                           'nRequiredIntermediateVariables'),
+            (fmi3InstanceEnvironment,            'instanceEnvironment'),
+            (fmi3CallbackLogMessageTYPE,         'logMessage'),
             (fmi3CallbackIntermediateUpdateTYPE, 'intermediateUpdate'),
-            (fmi3CallbackLockPreemptionTYPE, 'lockPreemption'),
-            (fmi3CallbackUnlockPreemptionTYPE, 'unlockPreemption'),
+            (fmi3CallbackLockPreemptionTYPE,     'lockPreemption'),
+            (fmi3CallbackUnlockPreemptionTYPE,   'unlockPreemption'),
         ], fmi3Instance)
 
         self._fmi3Function('fmi3FreeInstance', [(fmi3Instance, 'instance')], None)
@@ -345,6 +358,10 @@ class _FMU3(_FMU):
             (c_size_t,                    'nIntervals')
         ])
 
+        self._fmi3Function('fmi3EvaluateDiscreteStates', [
+            (fmi3Instance, 'instance')
+        ])
+
         self._fmi3Function('fmi3UpdateDiscreteStates', [
             (fmi3Instance,         'instance'),
             (POINTER(fmi3Boolean), 'discreteStatesNeedUpdate'),
@@ -505,7 +522,7 @@ class _FMU3(_FMU):
     def setDebugLogging(self, loggingOn, categories):
         categories_ = (fmi3String * len(categories))()
         categories_[:] = [c.encode('utf-8') for c in categories]
-        self.fmi3SetDebugLogging(self.component, fmi3True if loggingOn else fmi3False, len(categories), categories_)
+        self.fmi3SetDebugLogging(self.component, fmi3Boolean(loggingOn), len(categories), categories_)
 
     # Creation and destruction of FMU instances and setting debug status
 
@@ -517,20 +534,20 @@ class _FMU3(_FMU):
 
     def enterInitializationMode(self, tolerance=None, startTime=0.0, stopTime=None):
 
-        toleranceDefined = fmi3True if tolerance is not None else fmi3False
+        toleranceDefined = fmi3Boolean(tolerance is not None)
 
         if tolerance is None:
             tolerance = 0.0
 
-        stopTimeDefined = fmi3True if stopTime is not None else fmi3False
+        stopTimeDefined = fmi3Boolean(stopTime is not None)
 
         if stopTime is None:
             stopTime = 0.0
 
-        return self.fmi3EnterInitializationMode(self.component, toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime)
+        self.fmi3EnterInitializationMode(self.component, toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime)
 
     def exitInitializationMode(self):
-        return self.fmi3ExitInitializationMode(self.component)
+        self.fmi3ExitInitializationMode(self.component)
 
     # Clock related functions
 
@@ -558,11 +575,11 @@ class _FMU3(_FMU):
 
         return self.fmi3EnterEventMode(
             self.component,
-            fmi3True if stepEvent else fmi3False,
-            fmi3True if stateEvent else fmi3False,
+            fmi3Boolean(stepEvent),
+            fmi3Boolean(stateEvent),
             rootsFound,
             len(rootsFound),
-            fmi3True if timeEvent else fmi3False,
+            fmi3Boolean(timeEvent)
         )
 
     def updateDiscreteStates(self):
@@ -582,11 +599,11 @@ class _FMU3(_FMU):
                                       byref(nextEventTimeDefined),
                                       byref(nextEventTime))
 
-        return (discreteStatesNeedUpdate.value          != fmi3False,
-                terminateSimulation.value               != fmi3False,
-                nominalsOfContinuousStatesChanged.value != fmi3False,
-                valuesOfContinuousStatesChanged.value   != fmi3False,
-                nextEventTimeDefined.value              != fmi3False,
+        return (discreteStatesNeedUpdate.value,
+                terminateSimulation.value,
+                nominalsOfContinuousStatesChanged.value,
+                valuesOfContinuousStatesChanged.value,
+                nextEventTimeDefined.value,
                 nextEventTime.value)
 
     def terminate(self):
@@ -685,17 +702,21 @@ class _FMU3(_FMU):
         self.fmi3GetBoolean(self.component, vr, len(vr), value, nValues)
         return list(value)
 
-    def getString(self, vr):
+    def getString(self, vr, nValues=None):
+        if nValues is None:
+            nValues = len(vr)
         vr = (fmi3ValueReference * len(vr))(*vr)
-        value = (fmi3String * len(vr))()
-        self.fmi3GetString(self.component, vr, len(vr), value)
-        return list(value)
+        value = (fmi3String * nValues)()
+        self.fmi3GetString(self.component, vr, len(vr), value, nValues)
+        return list(map(lambda b: b.decode('utf-8'), value))
 
-    def getBinary(self, vr):
+    def getBinary(self, vr, nValues=None):
+        if nValues is None:
+            nValues = len(vr)
         vr = (fmi3ValueReference * len(vr))(*vr)
-        value = (fmi3Binary * len(vr))()
+        value = (fmi3Binary * nValues)()
         size = (c_size_t * len(vr))()
-        self.fmi3GetBinary(self.component, vr, len(vr), size, value, len(value))
+        self.fmi3GetBinary(self.component, vr, len(vr), size, value, nValues)
         return list(value)
 
     def getClock(self, vr, nValues=None):
@@ -871,7 +892,7 @@ class FMU3Model(_FMU3):
 
     def instantiate(self, visible=False, loggingOn=False):
 
-        resourceLocation = pathlib.Path(self.unzipDirectory, 'resources').as_uri()
+        resourcePath = os.path.join(self.unzipDirectory, 'resources') + os.path.sep
 
         # save callbacks from GC
         self.printLogMessage = fmi3CallbackLogMessageTYPE(printLogMessage)
@@ -879,9 +900,9 @@ class FMU3Model(_FMU3):
         self.component = self.fmi3InstantiateModelExchange(
             self.instanceName.encode('utf-8'),
             self.guid.encode('utf-8'),
-            resourceLocation.encode('utf-8'),
-            fmi3True if visible else fmi3False,
-            fmi3True if loggingOn else fmi3False,
+            resourcePath.encode('utf-8'),
+            fmi3Boolean(visible),
+            fmi3Boolean(loggingOn),
             fmi3InstanceEnvironment(),
             self.printLogMessage)
 
@@ -893,10 +914,10 @@ class FMU3Model(_FMU3):
     def enterContinuousTimeMode(self):
         return self.fmi3EnterContinuousTimeMode(self.component)
 
-    def completedIntegratorStep(self, noSetFMUStatePriorToCurrentPoint=fmi3True):
+    def completedIntegratorStep(self, noSetFMUStatePriorToCurrentPoint=True):
         enterEventMode = fmi3Boolean()
         terminateSimulation = fmi3Boolean()
-        self.fmi3CompletedIntegratorStep(self.component, noSetFMUStatePriorToCurrentPoint, byref(enterEventMode), byref(terminateSimulation))
+        self.fmi3CompletedIntegratorStep(self.component, fmi3Boolean(noSetFMUStatePriorToCurrentPoint), byref(enterEventMode), byref(terminateSimulation))
         return enterEventMode.value, terminateSimulation.value
 
     # Providing independent variables and re-initialization of caching
@@ -933,20 +954,20 @@ class FMU3Slave(_FMU3):
 
     def instantiate(self, visible=False, loggingOn=False, eventModeUsed=False, earlyReturnAllowed=False):
 
-        resourceLocation = pathlib.Path(self.unzipDirectory, 'resources').as_uri()
-
         # save callbacks from GC
         self.printLogMessage = fmi3CallbackLogMessageTYPE(printLogMessage)
         self.intermediateUpdate = fmi3CallbackIntermediateUpdateTYPE(intermediateUpdate)
 
+        resourcePath = os.path.join(self.unzipDirectory, 'resources') + os.path.sep
+
         self.component = self.fmi3InstantiateCoSimulation(
             self.instanceName.encode('utf-8'),
             self.guid.encode('utf-8'),
-            resourceLocation.encode('utf-8'),
-            fmi3True if visible else fmi3False,
-            fmi3True if loggingOn else fmi3False,
-            fmi3True if eventModeUsed else fmi3False,
-            fmi3True if earlyReturnAllowed else fmi3False,
+            resourcePath.encode('utf-8'),
+            fmi3Boolean(visible),
+            fmi3Boolean(loggingOn),
+            fmi3Boolean(eventModeUsed),
+            fmi3Boolean(earlyReturnAllowed),
             None, 0,
             fmi3InstanceEnvironment(),
             self.printLogMessage,
@@ -973,13 +994,23 @@ class FMU3Slave(_FMU3):
         self.fmi3GetOutputDerivatives(self.component, vr, len(vr), order, value)
         return list(value)
 
-    def doStep(self, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint=fmi3True):
+    def doStep(self, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint=True) -> Tuple[bool, bool, bool, float]:
+
         eventEncountered = fmi3Boolean()
         terminateSimulation = fmi3Boolean()
         earlyReturn = fmi3Boolean()
         lastSuccessfulTime = fmi3Float64()
-        status = self.fmi3DoStep(self.component, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint, byref(eventEncountered), byref(terminateSimulation), byref(earlyReturn), byref(lastSuccessfulTime))
-        return status, eventEncountered.value != fmi3False, terminateSimulation.value != fmi3False, earlyReturn.value != fmi3False, lastSuccessfulTime.value
+
+        self.fmi3DoStep(self.component,
+                        currentCommunicationPoint,
+                        communicationStepSize,
+                        fmi3Boolean(noSetFMUStatePriorToCurrentPoint),
+                        byref(eventEncountered),
+                        byref(terminateSimulation),
+                        byref(earlyReturn),
+                        byref(lastSuccessfulTime))
+
+        return eventEncountered.value, terminateSimulation.value, earlyReturn.value, lastSuccessfulTime.value
 
 
 class FMU3ScheduledExecution(_FMU3):
@@ -999,7 +1030,7 @@ class FMU3ScheduledExecution(_FMU3):
 
     def instantiate(self, visible=False, loggingOn=False):
 
-        resourceLocation = pathlib.Path(self.unzipDirectory, 'resources').as_uri()
+        resourcePath = os.path.join(self.unzipDirectory, 'resources') + os.path.sep
 
         def noop():
             pass
@@ -1013,9 +1044,9 @@ class FMU3ScheduledExecution(_FMU3):
         self.component = self.fmi3InstantiateScheduledExecution(
             self.instanceName.encode('utf-8'),
             self.guid.encode('utf-8'),
-            resourceLocation.encode('utf-8'),
-            fmi3True if visible else fmi3False,
-            fmi3True if loggingOn else fmi3False,
+            resourcePath.encode('utf-8'),
+            fmi3Boolean(visible),
+            fmi3Boolean(loggingOn),
             None, 0,
             fmi3InstanceEnvironment(),
             self.printLogMessage,
